-- ----------------------------------------------------------------------------
Library IEEE;
Use     IEEE.Std_Logic_1164.All;
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
Entity IdeInterface Is
    Port (
        AppleII_Clk_28MHz   : In  Std_Logic;  --mclk28
        AppleII_Reset       : In  Std_Logic;
        AppleII_SelectC7Io  : In  Std_Logic;
        CPU_ReadWrite       : In  Std_Logic; -- we_n
        CPU_DataIn          : In  Std_Logic_Vector(7 Downto 0); -- data_from_cpu[7:0]
        CPU_DataOut         : Out Std_Logic_Vector(7 Downto 0);
        CPU_Address         : In  Std_Logic_Vector(7 Downto 0);
        DiskMainCycle       : In  Std_Logic; --disk_main_cycyle

        IDE_SELECT  : Out   Std_Logic;
        IDE_POWER   : Out   Std_Logic;
        IDE_DRQ     : In    Std_Logic;
        IDE_RESET   : Out   Std_Logic;
        IDE_DATA_IO : InOut Std_Logic_Vector(15 Downto 0);
        IDE_IOR     : Out   Std_Logic;
        IDE_IOW     : Out   Std_Logic;
        IDE_A2      : Out   Std_Logic;
        IDE_A1      : Out   Std_Logic;
        IDE_A0      : Out   Std_Logic;
        IDE_CS1     : Out   Std_Logic;
        IDE_CS0     : Out   Std_Logic
    );
End;
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
Architecture RTL Of IdeInterface Is
-- ----------------------------------------------------------------------------
    Signal WriteEnable    : Std_Logic;
    Signal SelectHighByte : Std_Logic;
    Signal SelectLowByte  : Std_Logic;
    Signal SelectDrq      : Std_Logic;
    Signal IdeIOW         : Std_Logic;
    Signal HighByteRead   : Std_Logic_Vector(7 Downto 0);
    Signal HighByteWrite  : Std_Logic_Vector(7 Downto 0);
-- ----------------------------------------------------------------------------
Begin
-- ----------------------------------------------------------------------------
    WriteEnable <= Not CPU_ReadWrite;
    SelectHighByte <= '1' When ((CPU_Address(3 Downto 0) = "0000") And
                                 (AppleII_SelectC7Io='1'))
                          Else
                      '0';

    SelectLowByte <= '1' When ((CPU_Address(3 Downto 0) = "1000") And
                                (AppleII_SelectC7Io='1'))
                         Else
                     '0';
    SelectDrq <= '1' When (CPU_Address(3 Downto 0) = "0001")
                     Else
                 '0';

    IdeIOW <= Not (AppleII_SelectC7Io And Not SelectHighByte And     WriteEnable And DiskMainCycle);

    CPU_DataOut <= HighByteRead        When SelectHighByte='1' Else
                   IDE_DRQ & "0000000" When SelectDrq='1'      Else
                   IDE_DATA_IO(7 Downto 0);

    -- Addressing:
    -- 0xC0F0 - r/w MSB
    -- 0xC0F1 - read DRQ
    -- 0xC0F6 - alt status, dev cntrl
    -- 0xC0F8 - r/w data
    -- 0xC0F9 - error/features
    -- 0xC0FA - sector count
    -- 0xC0FB - sector N
    -- 0xC0FC - low cyl
    -- 0xC0FD - high cyl
    -- 0xC0FE - head
    -- 0xC0FF - status

    IDE_A2      <= CPU_Address(2);
    IDE_A1      <= CPU_Address(1);
    IDE_A0      <= CPU_Address(0);
    IDE_CS0     <= Not CPU_Address(3);
    IDE_CS1     <= CPU_Address(3) And Not SelectDrq;
    IDE_IOR     <= Not (AppleII_SelectC7Io And Not SelectHighByte And Not WriteEnable And DiskMainCycle);
    IDE_IOW     <= Not (AppleII_SelectC7Io And Not SelectHighByte And     WriteEnable And DiskMainCycle);
    IDE_DATA_IO <= (HighByteWrite & CPU_DataIn) When IdeIOW='0' Else (Others=>'Z'); --RAF: change from IdeIOW='1'
    IDE_SELECT  <= '0';
    IDE_POWER   <= '0';
    IDE_RESET   <= Not AppleII_Reset;

    Process(AppleII_Clk_28MHz)
    Begin
        If Rising_Edge(AppleII_Clk_28MHz) Then
            If (DiskMainCycle='1' And WriteEnable='0') Then
                If SelectLowByte='1' Then
                    HighByteRead <= IDE_DATA_IO(15 Downto 8);
                End If;
--                LowByteRead <= IDE_DATA_IO(7 Downto 0);
            End If;
        End If;
    End Process;

    Process(AppleII_Clk_28MHz)
    Begin
        If Rising_Edge(AppleII_Clk_28MHz) Then
            If (SelectHighByte='1' And DiskMainCycle='1' And WriteEnable='1') Then
                HighByteWrite <= CPU_DataIn;
            End If;
        End If;
    End Process;

-- ----------------------------------------------------------------------------
End RTL;
-- ----------------------------------------------------------------------------

