-- ----------------------------------------------------------------------------
Library IEEE;
Use     IEEE.Std_Logic_1164.All;
Use     IEEE.Std_Logic_Unsigned.All;
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
Entity KbdInterface Is
    Port (
        AppleII_Clk         : In    Std_Logic;
        AppleII_Reset       : In    Std_Logic;
        PS2_Data            : In    Std_Logic;
        PS2_Clk             : In    Std_Logic;
        KBD_Read            : In    Std_Logic;
        AppleII_ButtonLeft  : Out   Std_Logic;
        AppleII_ButtonRight : Out   Std_Logic;
        KBD_DataOut         : Out   Std_Logic_Vector(7 Downto 0);
        CPU_Reset           : Out   Std_Logic;
        KBD_ShiftLock       : Out   Std_Logic;
        KBD_Shift           : Out   Std_Logic;
        KBD_Available       : Out   Std_Logic
    );
End;
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
Architecture RTL Of KbdInterface Is
-- ----------------------------------------------------------------------------

    -- ------------------------------------------------------------------------
    Component PS2_Controller Is
        Port (
            Clk       : In  Std_Logic;
            Reset     : In  Std_Logic;
            PS2_CLK   : In  Std_Logic;
            PS2_Data  : In  Std_Logic;
            DoRead    : In  Std_Logic;
            Scan_Err  : Out Std_Logic;
            Scan_DAV  : Out Std_Logic;
            Scan_Code : Out Std_Logic_Vector(7 Downto 0)
        );
    End Component;
    -- ------------------------------------------------------------------------

    -- ------------------------------------------------------------------------
    Component KBD_Translate Is
        Port (
            Shift   : In  Std_logic;
            InCode  : In  Std_Logic_Vector(7 Downto 0);
            OutCode : Out Std_Logic_Vector(6 Downto 0)
        );
    End Component;
    -- ------------------------------------------------------------------------

    -- ------------------------------------------------------------------------
    Constant sState0 : Std_Logic_Vector(2 Downto 0) := "000";
    Constant sState1 : Std_Logic_Vector(2 Downto 0) := "001";
    Constant sState2 : Std_Logic_Vector(2 Downto 0) := "010";
    Constant sState3 : Std_Logic_Vector(2 Downto 0) := "011";
    Constant sState4 : Std_Logic_Vector(2 Downto 0) := "100";
    Constant sState5 : Std_Logic_Vector(2 Downto 0) := "101";
    Constant sState6 : Std_Logic_Vector(2 Downto 0) := "110";
    Constant sState7 : Std_Logic_Vector(2 Downto 0) := "111";
    Signal KbdState : Std_Logic_Vector(2 Downto 0);
    Signal KbdStateNext : Std_Logic_Vector(2 Downto 0);
    Signal ResetFromKey : Std_Logic;
    Signal Decoded1     : Std_Logic_Vector(6 Downto 0);
    Signal Reset_Delay  : Std_Logic_Vector(7 Downto 0);
    Signal Decoded      : Std_Logic_Vector(6 Downto 0);
    Signal ScanCode     : Std_Logic_Vector(7 Downto 0);
    Signal KbdAvailable : Std_Logic;
    Signal DoRead       : Std_Logic;
    Signal Scan_Err     : Std_Logic;
    Signal Scan_DAVi    : Std_Logic;
    Signal Ctrl         : Std_Logic;
    Signal Shift        : Std_Logic;
    Signal CodeLatched  : Std_Logic_Vector(7 Downto 0);
    Signal ButtonLeft   : Std_Logic;
    Signal ButtonRight  : Std_Logic;
    Signal ShiftLock    : Std_Logic;
    -- ------------------------------------------------------------------------


-- ----------------------------------------------------------------------------
Begin
-- ----------------------------------------------------------------------------


    PS2_Cntrl: PS2_Controller Port Map (
        Clk       => AppleII_Clk,
        Reset     => AppleII_Reset,
        PS2_CLK   => PS2_CLK,
        PS2_Data  => PS2_Data,
        DoRead    => DoRead,
        Scan_Err  => Scan_Err,
        Scan_DAV  => Scan_DAVi,
        Scan_Code => ScanCode
    );

    Code_Translate: KBD_Translate Port Map (
        Shift   => Shift,
        InCode  => CodeLatched,
        OutCode => Decoded
    );

    -- ------------------------------------------------------------------------
    CPU_Reset   <= AppleII_Reset Or ResetFromKey;
--    Decoded1    <= (Decoded And "1011111") When ((Decoded(6 Downto 5) = "11") And ShiftLock='1') Else Decoded;
    Decoded1    <= Decoded(6) & '0' & Decoded(4 Downto 0) When ((Decoded(6 Downto 5) = "11") And ShiftLock='1') Else Decoded;
    KBD_DataOut <= KbdAvailable & "00" & Decoded(4 Downto 0) When Ctrl='1' Else KbdAvailable & Decoded1;

    DoRead              <= Scan_DAVi;
--    PS2_CLK             <= 'Z';
--    PS2_Data            <= 'Z';
    AppleII_ButtonLeft  <= ButtonLeft;
    AppleII_ButtonRight <= ButtonRight;
    KBD_ShiftLock       <= ShiftLock;
    KBD_Shift           <= Shift;
    KBD_Available       <= KbdAvailable;

    -- ------------------------------------------------------------------------

    -- ------------------------------------------------------------------------
    Process(AppleII_Clk)
    Begin
        If Rising_Edge(AppleII_Clk) Then
            If (Reset_Delay /= x"FF") Then
                Reset_Delay <= Reset_Delay + 1;
            End If;
        End If;
    End Process;
    -- ------------------------------------------------------------------------

    -- ------------------------------------------------------------------------
    Process(AppleII_Clk)
    Begin
        If Rising_Edge(AppleII_Clk) Then
            If AppleII_Reset='1' Then
                Shift     <= '0';
                Ctrl      <= '0';
                ShiftLock <= '0';
            Else
                If (KbdState=sState1) Then
                       If (ScanCode = x"12") Or
                          (ScanCode = x"59") Then Shift        <= '1';
                    ElsIf (ScanCode = x"14") Then Ctrl         <= '1';
                    ElsIf (ScanCode = x"1F") Then ButtonLeft   <= '1';
                    ElsIf (ScanCode = x"27") Then ButtonRight  <= '1';
                    ElsIf (ScanCode = x"58") Then ShiftLock    <= Not ShiftLock;
                    ElsIf (ScanCode = x"7E") Then ResetFromKey <= '1';
                    End If;
                Else
                    If (KbdState = sState6) Then
                           If (ScanCode = x"12") Or
                              (ScanCode = x"59") Then Shift        <= '0';
                        ElsIf (ScanCode = x"14") Then Ctrl         <= '0';
                        ElsIf (ScanCode = x"1F") Then ButtonLeft   <= '0';
                        ElsIf (ScanCode = x"27") Then ButtonRight  <= '0';
                        ElsIf (ScanCode = x"7E") Then ResetFromKey <= '0';
                        End If;
                    End If;
                End If;
            End If;
        End If;
    End Process;
    -- ------------------------------------------------------------------------

    -- ------------------------------------------------------------------------
    Process(AppleII_Clk)
    Begin
        If Rising_Edge(AppleII_Clk) Then
            If (AppleII_Reset='1') Then
                KbdState     <= sState0;
                CodeLatched  <= (Others=>'0');
                KbdAvailable <= '0';
            Else
                KbdState <= KbdStateNext;
                If (KBD_Read='1') Then
                    KbdAvailable <= '0';
                End If;
                If (KbdState = sState7) Then
                    CodeLatched  <= ScanCode;
                    KbdAvailable <= '1';
                End If;
            End If;
        End If;
    End Process;
    -- ------------------------------------------------------------------------

    -- ------------------------------------------------------------------------
    Process(KbdState, ScanCode, Scan_DAVi)
    Begin
        KbdStateNext <= KbdState;

        Case (KbdState) Is
            When sState0 =>
                 If (Scan_DAVi='1') Then
                    KbdStateNext <= sState1;
                 End If;
            When sState1 => KbdStateNext <= sState2;
            When sState2 =>
                    If (ScanCode=x"F0")                                         Then KbdStateNext <= sState3;
                 ElsIf (ScanCode=x"E0")                                         Then KbdStateNext <= sState0;
                 ElsIf (ScanCode=x"12") Or (ScanCode=x"14") Or (ScanCode=x"59") Then KbdStateNext <= sState0;
                                                                                Else KbdStateNext <= sState7;
                 End If;

            When sState3 =>
                 KbdStateNext <= sState4;
            When sState4 =>
                 KbdStateNext <= sState5;
            When sState5 =>
                 If Scan_DAVi='1' Then
                    KbdStateNext <= sState6;
                 End If;
            When sState6 =>
                 KbdStateNext <= sState0;
            When sState7 =>
                 KbdStateNext <= sState0;
            When Others  =>
                 KbdStateNext <= sState0;
        End Case;
    End Process;
    -- ------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
End RTL;
-- ----------------------------------------------------------------------------
